package infomat.algorithms.labeling;

import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import infomat.vectorspace.vector.*;

import mro.util.CurrentClassGetter;
import mro.util.properties.*;

import java.util.*;


/**
 * Reduces IObjectGroups and IObjectGroupings using their ILabels.
 */
public class LabelReducer implements HasProperties{
    mro.util.properties.Properties properties;

    public LabelReducer(){
	this.properties = getDefaultProperties();
    }

    public LabelReducer(mro.util.properties.Properties properties){
	this.properties = properties;
    }

    public mro.util.properties.Properties getDefaultProperties(){

	mro.util.properties.Properties properties = new mro.util.properties.Properties("Label Reducer Properties", CurrentClassGetter.getClassName());
		
	try{

	    Property p = new Property("Reduction Method", 2, "The reduction method");

	    PropertyValueDescription pvd = new PropertyValueDescription("1", "At least \"Number of shared objects\" in common between ILabel and IObject.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription("2", "Similarity to the ILabel considered as a centroid. See \"Bigger similarity to ILabel centroid than\".");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	    p = new Property("Number of shared objects", 1, "The number of objects that has to be shared between the ILabel and the IObjects");
	    properties.addProperty(p);

	    p = new Property("Bigger similarity to ILabel centroid than", 0.0, "IObjects with bigger similarity to the ILabel centroid are saved.");
	    properties.addProperty(p);


	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}

	return properties;
    }

    public Object getPropertyValue(String name){
	return properties.getPropertyValue(name);
    }

    public void setProperty(String name, Object value) 
	throws PropertyNotFoundException{
	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
	return properties;
    }

    public void setProperties(mro.util.properties.Properties properties){
	this.properties=properties;
    }


    /**
     * Returns two IObjectGrouping:s: 
     * 0) the reduced grouping, 1) the rest grouping.
     */
    public IObjectGrouping[] reduceGroupingByLabeling(IObjectGrouping ioging, HashMap<IObjectGroup, ILabel> labeling, IMatrix matrix, SparseISimilarity similarity){

	int reductionMethod = 
	    (Integer) properties.getPropertyValue("Reduction Method");

	boolean row=true;
	IObjectSet ioset = matrix.getRowIObjectSet();
	if(!ioset.containsIObjectGrouping(ioging)){
	    ioset = matrix.getColumnIObjectSet();
	    row=false;
	}

	Vector<IObjectGroup> newGroups = new Vector<IObjectGroup>();
	Vector<IObjectGroup> restGroups = new Vector<IObjectGroup>();
	Vector<IObjectGroup> groups = ioging.getGroups();
	for(int i=0; i<groups.size(); i++){
	    IObjectGroup iog = groups.elementAt(i);
	    ILabel label = labeling.get(iog);
	    IObjectGroup[] newIOGs = null;
	    if(reductionMethod==1)
		newIOGs = reduceGroupBySharedLabelObjects(iog, label, matrix, row);
	    else if(reductionMethod==2)
		newIOGs = reduceGroupByLabelCentroid(iog, label, similarity);

	    newGroups.add(newIOGs[0]);
	    restGroups.add(newIOGs[1]);

	    /*
	    IObjectGroup newIOG = null;
	    if(reductionMethod==1)
		newIOG = reduceGroupBySharedLabelObjects(iog, label, matrix);
	    else if(reductionMethod==2)
		newIOG = reduceGroupByLabelCentroid(iog, label, matrix);
	    // Hr mrklig ndring 080512
	    newGroups.add(newIOG);
	    */
	}


	String name = "Label Reduced "+ioging.getName();
	IObjectGrouping labelIOGING = new IObjectGrouping(name);
	int i=1;
	while(!ioset.setGrouping(labelIOGING)){
	    labelIOGING.setName(name+" ("+i+")");
	    i++;
	}
	labelIOGING.addAllGroups(newGroups);

	name = "Rest Label Reduced "+ioging.getName();
	IObjectGrouping restIOGING = new IObjectGrouping(name);
	i=1;
	while(!ioset.setGrouping(restIOGING)){
	    restIOGING.setName(name+" ("+i+")");
	    i++;
	}
	restIOGING.addAllGroups(restGroups);

	IObjectGrouping[] ret = new IObjectGrouping[2];
	ret[0] = labelIOGING;
	ret[1] = restIOGING;

	return ret;
    }


    /**
     * Makes a new (reduced) IObjectGroup 
     * by considering the ILabel a centroid and
     * removing those IObjects with similarity higer than a parameter.
     * Returns two IObjectGroup:s: 
     * 0) the one described, 1) the rest.
     */
    public IObjectGroup[] reduceGroupByLabelCentroid(IObjectGroup iog, ILabel label, SparseISimilarity similarity){

	double leastSim = (Double) properties.getPropertyValue("Bigger similarity to ILabel centroid than");

	IObjectGroup newGroup = 
	    new IObjectGroup("Label Reduced "+iog.getName());
	IObjectGroup restGroup = 
	    new IObjectGroup("Rest Label Reduced "+iog.getName());

	label.sortInOrderValueOrder();
	HashMap<IObject, Float> labelCentroid = 
	    label.getObjectOrderValueMap();
	IVector<IObject> labelVec = new IVector<IObject>(labelCentroid);

	Vector<IObject> ordered = iog.getOrdered();
	Vector<IObject> newVec = new Vector<IObject>();
	for(int i=0; i<ordered.size(); i++){
	    IObject io = ordered.elementAt(i);
	    IVector<IVectorIndex> repr = similarity.getRepresentation(io);
	    float sim = similarity.sim(repr, labelVec);

	    /*
	    HashMap<IObject, Float> representation = 
		io.getRepresentation(matrix);
	    float sim = 
		Sorter.compareObjectToCentroid(representation, labelCentroid);
	    */
	    //	    System.out.println("sim: "+sim);

	    if(sim>leastSim)
		newGroup.add(io, sim);
	    else
		restGroup.add(io, sim);
	    /*
	    if(sim>leastSim)
		newVec.add(io, sim);
	    else
		restVec.add(io, sim);
	    */
	}

	/*
	IObjectGroup newGroup = 
	    new IObjectGroup("Label Reduced "+iog.getName());
	newGroup.addAll(newVec);

	IObjectGroup restGroup = 
	    new IObjectGroup("Rest Label Reduced "+iog.getName());
	restGroup.addAll(newVec);
	*/

	IObjectGroup[] ret = new IObjectGroup[2];
	ret[0] = newGroup;
	ret[1] = restGroup;

	return ret;
    }
    
    /**
     * Makes a new (reduced) IObjectGroup that consist of 
     * those IObject:s that contain 
     * the IObject:s of the ILabel in their representation.
     * Returns two IObjectGroup:s: 
     * 0) the one described, 1) the rest.
     */
    public IObjectGroup[] reduceGroupBySharedLabelObjects(IObjectGroup iog, ILabel label, IMatrix matrix, boolean row){
	int numberOfShared = 
	    (Integer) properties.getPropertyValue("Number of shared objects");

	IObjectGroup newGroup = 
	    new IObjectGroup("Label Reduced "+iog.getName());
	IObjectGroup restGroup = 
	    new IObjectGroup("Rest Label Reduced "+iog.getName());

	Vector<IObject> ordered = iog.getOrdered();
	Vector<IObject> newVec = new Vector<IObject>();
	Vector<IObject> restVec = new Vector<IObject>();
	for(int i=0; i<ordered.size(); i++){
	    IObject io = ordered.elementAt(i);
	    //	    HashSet<IObject> ioset = matrix.getIObjectsForIObject(io);

	    HashSet<IObject> ioset = matrix.getOtherIObjectsForIObject(io);

	    Vector<IObject> labelIOs = label.getOrdered();
	    int count=0;
	    for(int j=0; j<labelIOs.size(); j++){
		if(ioset.contains(labelIOs.elementAt(j))){
		    count++;
		    if(count>=numberOfShared){
			break;
		    }
		}
	    }
	    if(count>=numberOfShared)
		newGroup.add(io, count);
	    else
		restGroup.add(io, count);
	    /*
	    if(count>=numberOfShared)
		newVec.add(io, count);
	    else
		restVec.add(io, sim);
	    */
	}

	/*
	IObjectGroup newGroup = 
	    new IObjectGroup("Label Reduced "+iog.getName());
	newGroup.addAll(newVec);

	IObjectGroup restGroup = 
	    new IObjectGroup("Rest Label Reduced "+iog.getName());
	restGroup.addAll(newVec);
	*/

	IObjectGroup[] ret = new IObjectGroup[2];
	ret[0] = newGroup;
	ret[1] = restGroup;

	return ret;
    }


}
