package infomat.vectorspace;

import infomat.vectorspace.vector.*;

import mro.util.measures.*;

import java.util.*;


/**
 * A IObjectGroup is a set of IObject:s.
 *
 * @param ordered A vector of the IObject:s.
 * @param objectOrderIndexMap A mapping from IObject to index in the vector "ordered".
 * @param objectOrderValueMap A mapping from IObject to a value. These values are the basis of the ordering in the vector "ordered".
 *
 * @version          070314
 * @author           Magnus Rosell
 */
public class IObjectGroup extends IObject {
    public static int maxIdSoFar=0;

    Vector<IObject> ordered; // The ordered vector
    public HashMap<IObject, Integer> objectOrderIndexMap; // The vector index
    HashMap<IObject, Float> objectOrderValueMap; // The basis for the ordering
	
    public IObjectGroup(String name){ 
	super(name, 1, ++maxIdSoFar); // 1 - IObjectGroup

	ordered = new Vector<IObject>();
	objectOrderIndexMap = new HashMap<IObject, Integer>();
	objectOrderValueMap = new HashMap<IObject, Float>();
    }

    // 080724
    /**
     * Returns the part of the group that starts with fromObject and
     * ends with toObject, in the same oreder as this group.
     * If fromObject lies after toObject they are inverted - that
     * is the method always returns a group ordered in the same way
     * as the original.
     */
    public IObjectGroup subGroup(IObject fromObject, IObject toObject){
	IObjectGroup ret = new IObjectGroup("Part of"+getString());
	Integer from = objectOrderIndexMap.get(fromObject);
	Integer to = objectOrderIndexMap.get(toObject);
	if(from==null || to==null)
	    return ret;
	if(from>to){
	    int tmp = from;
	    from = to;
	    to = tmp;
	}
	/*
	else if(from==to){
	    ret.add(fromObject, objectOrderValueMap.get(fromObject));
	    return ret;
	}
	*/
	for(int i=from; i<=to; i++){
	    IObject io = ordered.get(i);
	    ret.add(io, objectOrderValueMap.get(io));
	}
	return ret;
    }

    public IObjectGroup(String name, int id){ 
	super(name, 1, id); // 1 - IObjectGroup
	if(id>maxIdSoFar)
	    maxIdSoFar=id;

	ordered = new Vector<IObject>();
	objectOrderIndexMap = new HashMap<IObject, Integer>();
	objectOrderValueMap = new HashMap<IObject, Float>();
    }

    /**
     * Used by IObjectGrouping and IObjectSet.
     */
    protected IObjectGroup(String name, int kind, int id){ 
	super(name, kind, id); 
	ordered = new Vector<IObject>();
	objectOrderIndexMap = new HashMap<IObject, Integer>();
	objectOrderValueMap = new HashMap<IObject, Float>();
    }

    /**
     * Copy constructor. Makes new instances of everything, 
     * but the IObject:s, which are the same.
     * Gives the new IObjectGroup the same name, 
     * but a different id.
     */
    public IObjectGroup(IObjectGroup iog){
	super(iog.getName(), 1, ++maxIdSoFar);
	ordered = new Vector<IObject>();
	objectOrderIndexMap = new HashMap<IObject, Integer>();
	objectOrderValueMap = new HashMap<IObject, Float>();

	// IObjects
	Iterator<Map.Entry<IObject, Float>> it = 
	    iog.getObjectOrderValueMap().entrySet().iterator();
	while(it.hasNext()){
	    Map.Entry<IObject, Float> entry = it.next();
	    add(entry.getKey(), entry.getValue());
	}
    }


    /**
     * Copies references.
     */
    /*
      public IObjectGroup(IObjectGroup cog){
      this.ordered = cog.ordered;
      this.objectOrderIndexMap=cog.objectOrderIndexMap;
      this.objectOrderValueMap=cog.objectOrderValueMap;
      }
    */

    /**
     * Adds a new IObject in the place the value implicates.
     * For the object to get the right place the group has to be
     * sorted through {@link #sortInOrderValueOrder()}.
     */
    public void add(IObject co, float orderValue){
	objectOrderIndexMap.put(co, ordered.size());
	ordered.add(co);
	objectOrderValueMap.put(co, orderValue);
    }

    /**
     * Adds a new IObject at the next place.
     */
    public void add(IObject co){
	add(co, (float) ordered.size());
    }

    /**
     * Adds all the IObject:s at the end.
     */
    public void addAll(Vector<IObject> iobjects){
	float orderValue = 0.0f;
	if(ordered.size()>0){
	    IObject io = ordered.lastElement();
	    orderValue = objectOrderValueMap.get(io);
	}
	    
	for(int i=0; i<iobjects.size(); i++)
	    add(iobjects.elementAt(i), ++orderValue);
    }

    public void addAll(IVector<IObject> ivec){
	Vector<IObject> vec = ivec.getUnorderedVector();
	for(int i=0; i<vec.size(); i++){
	    IObject io = vec.get(i);
	    add(io, ivec.getValue(io));
	}
    }


    /**
     * Reorders the IObjects randomly.
     */
    public void randomOrder(){
	Vector<IObject> tmp = ordered;
	Collections.shuffle(tmp);
	removeAll();
	addAll(tmp);
    }


    /**
     * Removes the IObject.
     */
    public void remove(IObject co){
	if(contains(co)){
	    ordered.remove(co);
	    objectOrderIndexMap.remove(co);
	    objectOrderValueMap.remove(co);
	}
    }

    public void remove(IVector<IObject> ivec){
	Vector<IObject> vec = ivec.getUnorderedVector();
	for(int i=0; i<vec.size(); i++)
	    remove(vec.get(i));
    }

    public void removeAll(){
	ordered = new Vector<IObject>();
	objectOrderIndexMap = new HashMap<IObject, Integer>();
	objectOrderValueMap = new HashMap<IObject, Float>();    	
    }

    /**
     * Removes IObject:s that are not in the given IObjectGroup.
     * Does not check for the opposite: objects that are in this, but not
     * in the given.
     */
    public void removeObjectsNotIn(IObjectGroup group){
	Vector<IObject> removed = new Vector<IObject>();
	Vector<IObject> ordered = getOrdered();
	for(int i=0; i<ordered.size(); i++){
	    IObject obj = ordered.elementAt(i);
	    if(!group.contains(obj))
		removed.add(obj);
	}
	for(int i=0; i<removed.size(); i++){
	    remove(removed.elementAt(i));
	}
	sortInOrderValueOrder();
    }

    /**
     * Returns a IObjectGroup with the IObject:s in this 
     * that is not in the specified IObjectGroup.
     */ 
    public IObjectGroup notIn(IObjectGroup group){
	IObjectGroup ret = 
	    new IObjectGroup(getName()+" minus "+group.getName());
	Vector<IObject> ordered = getOrdered();
	for(int i=0; i<ordered.size(); i++){
	    IObject obj = ordered.elementAt(i);
	    if(!group.contains(obj))
		ret.add(obj);
	}
	return ret;
    }

    /**
     * Returns true when the IObjectGroup contains the IObject.
     * False otherwise.
     */
    public boolean contains(IObject co){
	return objectOrderIndexMap.containsKey(co);
    }

    /**
     * Returns true if the IObjectGroup contains these IObject:s
     * and no other.
     */
    public boolean containsPrecisely(Vector<IObject> objs){
	if(objs.size() != numberOfIObjects())
	    return false;
	
	for(int i=0; i<objs.size(); i++){
	    if(!contains(objs.elementAt(i)))
		return false;
	}
	return true;
    }

    /**
     * Returns the number of IObjects in the IObjectGroup.
     */
    public int numberOfIObjects(){
	return ordered.size();
    }

    /**
     * Returns the name of thie IObjectGroup.
     */
    public String getName(){
	return super.getString();
    }

    public void setName(String name){
	super.setString(name);
    }

    /**
     * Returns the IObject at the i:the place according to the ordering.
     * @see #sortInOrderValueOrder
     */
    public IObject getIObjectAt(int i){
	return ordered.elementAt(i);
    }

    /**
     * Returns the order number of the IObject.
     * If it is not present -1 is returned.
     * @see #sortInOrderValueOrder
     */
    public int getIObjectOrderIndex(IObject co){
	/*
	  System.out.println("IObjectGroup - objectOrderIndexMap: "+
	  //model.getIMatrix().
	  objectOrderIndexMap);
	*/
	Integer i = objectOrderIndexMap.get(co);
	if(i==null)
	    return -1;
	return i;
	//		return objectOrderIndexMap.get(co);
    }

    /**
     * Returns a vector of the ordered IObjects.
     * @see #sortInOrderValueOrder
     */
    public Vector<IObject> getOrdered(){
	return ordered;
    }

    public IVector<IObject> getIVector(){
	IVector<IObject> ret = new IVector<IObject>(objectOrderValueMap);
	return ret;
    }

    /**
     * Sets this group to contain the IObject:s in the specified IVector.
     * Use with caution!
     */
    public void setIVector(IVector<IObject> vec){
	removeAll();
	addAll(vec);
    }


    /**
     * Returns a HashMap with the objects mapped to the values that
     * are the basis for the ordering.
     * @return a HashMap
     */
    public HashMap<IObject, Float> getObjectOrderValueMap(){
	return objectOrderValueMap;
    }
	
	
	
    /** 
     * Returns a sorted centroid vector
     */
    public Vector<Map.Entry<IObject, Float>> getSortedObjectOrderValueVector(){
    	
    	Vector<Map.Entry<IObject, Float>> sorted= new Vector<Map.Entry<IObject, Float>>(getObjectOrderValueMap().entrySet());
    	Collections.sort(sorted, new MapEntryComparator());
    	
    	return sorted;
    }

    
    public void sortInCurrentVectorOrderWithInvertedOrderNumberAsValue(){
	Vector<IObject> vec = getOrdered();
	removeAll();
	for(int i=0; i<vec.size(); i++){
	    add(vec.get(i), vec.size()-i);
	}
    }

    

	
    /**
     * Sorts the IObjects in the vector "ordered" 
     * according to the values they have.
     */
    public void sortInOrderValueOrder(){
	IObjectComparator coc = new IObjectComparator(this);
	Collections.sort(ordered, coc);

	objectOrderIndexMap = new HashMap<IObject, Integer>();
	for(int i=0; i<ordered.size(); i++){
	    IObject co = ordered.elementAt(i);
	    objectOrderIndexMap.put(co, i);
	}
    }

    /**
     * Returns the number of IObjects that are shared between the
     * this and another IObjectGroup.
     */
    public int numberOfSharedIObjects(IObjectGroup iog){
    	
	Set<IObject> s = (Set<IObject>)
	    objectOrderIndexMap.keySet();
	Set<IObject> ss = (Set<IObject>)
	    iog.objectOrderIndexMap.keySet();
	
	HashSet<IObject> hs;
	Set<IObject> set;
	if(s.size()>ss.size()){
	    hs = new HashSet<IObject>(s);
	    set = ss;
	}
	else{
	    hs = new HashSet<IObject>(ss);
	    set = s;
	}
	
	int n = 0;
	Iterator<IObject> i = set.iterator();
	while(i.hasNext()){
	    if(hs.contains(i.next())){
		n++;
	    }
	}
	return n;
    }
    
    
    // gra sorterad centroid hr
    
    /** Returns a sorted centroid vector
     */

    /*    
    public Vector<Map.Entry<IObject, Float>> getSortedCentroid(IMatrix matrix){
    	
    	Vector<Map.Entry<IObject, Float>> sortedCentroid = new Vector<Map.Entry<IObject, Float>>(getRepresentation(matrix).entrySet());
    	Collections.sort(sortedCentroid, new MapEntryComparator());
    	
    	return sortedCentroid;
    }
    */

    /**
     * Returns a centroid in the form of a HashMap, mapping
     * IObject to a Float value.
     */
    /*
    public HashMap<IObject, Float> getRepresentation(IMatrix matrix){
	HashMap<IObject, Float> centroid = new HashMap<IObject, Float>();
	for(int i=0; i<ordered.size(); i++){
	    IObject co = ordered.elementAt(i);
	    Vector features = matrix.getCellsWith(co);
	    if(features!=null){
		for(int f=0; f<features.size(); f++){
		    IMatrixCell sc = (IMatrixCell) features.elementAt(f);
		    IObject feature = sc.getRowIObject();
		    if(co==feature)
			feature = sc.getColumnIObject();

		    Float w = centroid.get(feature);
		    if(w!=null)
			centroid.put(feature, w+sc.getWeight());
		    else
			centroid.put(feature, sc.getWeight());
		}
	    }
	}

	Iterator<IObject> it = centroid.keySet().iterator();
	while(it.hasNext()){
	    IObject key = it.next();
	    centroid.put(key, centroid.get(key)/numberOfIObjects());
	}
	return centroid;
    }
    */

    /**
     * A private class that is used to order the IObjects.
     * The order depends on the value in the "objectOrderValueMap"
     * parameter of the IObjectGroup. If the two IObject:s are equal
     * in "objectOrder" they are sorted lexiographically.
     */
    private class IObjectComparator implements Comparator{

	private IObjectGroup cog;

	public IObjectComparator(IObjectGroup cog){
	    this.cog=cog;
	}

	/**
	 * Compares two IObject:s and returns 1, 0 or -1 
	 *
	 * @param  o1 a IObject
	 * @param  o2 a IObject
	 * @return int 
	 */
	public int compare(Object o1, Object o2) {
	    float value1 = cog.objectOrderValueMap.get((IObject) o1);
	    float value2 = cog.objectOrderValueMap.get((IObject) o2);

	    // Strst frst
	    if(value1<value2)
		return 1;
	    else if (value1>value2)
		return -1;
	    else{
		String s1 = ((IObject) o1).getString();
		String s2 = ((IObject) o2).getString();
		return s1.compareTo(s2);
		//		return 0;
	    }
	}    

    }
	
    /**
     * A private class that is used to order the MapEntry:s.
     * @see #getSortedCentroid
     */
    private class MapEntryComparator implements Comparator{

	/**
	 * Compares two IObject:s and returns 1, 0 or -1 
	 *
	 * @param  o1 a IObject
	 * @param  o2 a IObject
	 * @return int 
	 */
	public int compare(Object o1, Object o2) {
	    float value1 = (Float)((Map.Entry)o1).getValue();
	    float value2 = (Float)((Map.Entry)o2).getValue();
	    // Strst frst
	    if(value1<value2)
		return 1;
	    else if (value1>value2)
		return -1;
	    else
		return 0;
	}    

    }



}

