package infomat.vectorspace;

import infomat.vectorspace.vector.*;

import java.util.*;

/**
 * A IObjectGrouping is a set of IObjectGroup:s.
 * It is ordered in a Vector. The order is deemed important.
 * OBS. A IObjectGrouping is only implicitely connected to an IObjectSet.
 * If the IObjectSet is altered this is not reflected in the IObjectGrouping.
 *
 * @param groupOrdered A (ordered) vector of the IObjectGroup:s.
 *
 * @version          070314
 * @author           Magnus Rosell
 */
public class IObjectGrouping extends IObjectGroup
{
    public static int groupingMaxIdSoFar=0;

    Vector<IObjectGroup> groupOrdered;

    public IObjectGrouping(String groupingName, int groupingId){
	super(groupingName, 2, groupingId); // 2 - IObjectGrouping
	//this.groupingName=groupingName;
	//this.groupingId=groupingId;
	if(groupingId>groupingMaxIdSoFar)
	    groupingMaxIdSoFar=groupingId;
	groupOrdered = new Vector<IObjectGroup>();
    }

    public IObjectGrouping(String groupingName){
	super(groupingName, 2, ++groupingMaxIdSoFar);
	//this.groupingName=groupingName;
	//this.groupingId=++groupingMaxIdSoFar;
	groupOrdered = new Vector<IObjectGroup>();
    }

    /**
     * Copy constructor. Makes new instances of everything, 
     * but the IObject:s, which are the same.
     * Gives the new IObjectGrouping the same name, 
     * but a different id.
     */
    public IObjectGrouping(IObjectGrouping ioging){
	//	super((IObjectGroup) ioging, 2, ++groupingMaxIdSoFar);
	super(ioging.getName(), 2, ++groupingMaxIdSoFar);

	// IObjects
	Iterator<Map.Entry<IObject, Float>> it = 
	    ioging.getObjectOrderValueMap().entrySet().iterator();
	while(it.hasNext()){
	    Map.Entry<IObject, Float> entry = it.next();
	    add(entry.getKey(), entry.getValue());
	}

	// Groups
	groupOrdered = new Vector<IObjectGroup>();

	Vector<IObjectGroup> groups = ioging.getGroups();
	for(int i=0; i<groups.size(); i++)
	    groupOrdered.add(new IObjectGroup(groups.elementAt(i)));
    }



    /**
     * Adds a IObjectGroup at the next place.
     */
    public void add(IObjectGroup cog){
	groupOrdered.add(cog);
	Vector<IObject> objs = cog.getOrdered();
	for(int i=0; i<objs.size(); i++)
	    add(objs.elementAt(i));
    }
	
    /**
     * Adds all IObjectGroup:s at the next places.
     */
    public void addAllGroups(Vector<IObjectGroup> vec){
	for(int i=0; i<vec.size(); i++)
	    add(vec.elementAt(i));
    }

    /**
     * Removes a IObjectGroup.
     */
    public void remove(IObjectGroup cog){
	groupOrdered.remove(cog);
	Vector<IObject> objs = cog.getOrdered();
	for(int i=0; i<objs.size(); i++)
	    remove(objs.elementAt(i));
    }

    /**
     * Removes all IObjectGroups.
     */
    protected void removeAllGroups(){
	removeAll();
	groupOrdered = new Vector<IObjectGroup>();
    }

    /**
     * Removes all empty IObjectGroup:s. 
     * Some algorithms may produce empty groups.
     * The interface does not handle empty groups.
     */
    public void removeEmptyGroups(){
	Vector<IObjectGroup> vec = new Vector<IObjectGroup>();
	for(int i=0; i<groupOrdered.size(); i++)
	    if(groupOrdered.elementAt(i).numberOfIObjects()>0)
		vec.add(groupOrdered.elementAt(i));
	groupOrdered=vec;
    }
	
	
    /**
     * Removes the IObject from all the IObjectGroup:s of the grouping.
     * If any of these becomes empty they are removed.
     */
    public void remove(IObject co){
	super.remove(co);
	//	System.out.println("Number of groups: "+groupOrdered.size());
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup cog = groupOrdered.elementAt(i);
	    cog.remove(co);
	    if(cog.numberOfIObjects()<1){
		//System.out.println("JA. Tar bort!!!!!!!!!!!!!!");
		remove(cog);
		//System.out.println("Number of groups: "+groupOrdered.size());
	    }
	}
    }

    public void remove(IVector<IObject> ivec){
	Vector<IObjectGroup> groups = getGroups();
	Vector<IObject> vec = ivec.getUnorderedVector();
	for(int i=0; i<vec.size(); i++){
	    IObject io = vec.get(i);
	    remove(io);
	    for(int j=0; j<groups.size(); j++){
		groups.get(j).remove(ivec);
	    }
	}
	sort();
    }


    /**
     * Returns the ordered IObjectGroup:s in a vector.
     */
    public Vector<IObjectGroup> getGroups(){
	return groupOrdered;
    }

    /**
     *
     */
    public boolean contains(IObjectGroup group){
	return groupOrdered.contains(group);
    }

	
    /**
     * Removes all groups and adds the vector.
     * Use this method with caution! All groups are removed!!
     */
    public void setGroups(Vector<IObjectGroup> iog){
	removeAllGroups();
	addAllGroups(iog);
    }

    /**
     * Reorders the groups according to a mapping in the form of
     * an int array. Does not do any check of the mapping.
     */
    public void reorder(int[] mapping){
	Vector<IObjectGroup> newOrdered = new Vector<IObjectGroup>();
	for(int i =0; i<mapping.length; i++){
	    newOrdered.add(groupOrdered.elementAt(mapping[i]));
	}
	removeAllGroups();
	addAllGroups(newOrdered);
	sort();
    }    

    /**
     * Returns the the number of IObjectGroup:s in the IObjectGrouping. 
     */
    public int numberOfIObjectGroups(){
	return groupOrdered.size();
    }

    /**
     * Returns a vector of all the IObjects contained in this IObjectGrouping
     * in the order suggested by the ordering of the IObjectGroup:s and
     * their internal order.
     */
    public Vector<IObject> getGroupingVector(){
	Vector<IObject> vec = new Vector<IObject>();
	for(int i=0; i<groupOrdered.size(); i++){
	    vec.addAll(groupOrdered.elementAt(i).getOrdered());
	}
	return vec;
    }

    /** 
     * Returns the IObjectGroup in this IObjectGrouping 
     * that contains this IObject.
     */
    public IObjectGroup inWhichIObjectGroup(IObject co){
	IObjectGroup cog=null;
	for(int i=0; i<groupOrdered.size(); i++){
	    cog=groupOrdered.elementAt(i);
	    if(cog.contains(co))
		return cog;
	}
	return null;
    }
	
    /** 
     * Returns the number of the IObjectGroup in this IObjectGrouping 
     * that contains this IObject.
     */
    public int inWhichIObjectGroupOrdered(IObject co){
	IObjectGroup cog=null;
	for(int i=0; i<groupOrdered.size(); i++){
	    cog=groupOrdered.elementAt(i);
	    if(cog.contains(co))
		return i;
	}
	return -1;
    }
	
    /**
     * Returns true if this IObjectGrouping contains the IObjectGroup.
     */
    public boolean containsIObjectGroup(IObjectGroup iog){
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup group=groupOrdered.elementAt(i);
	    if(group==iog)
		return true;
	}
	return false;
    }


    /**
     * Sorts all IObjectGroup:s in their internal order.
     * @see IObjectGroup#sortInOrderValueOrder
     */
    /*
      public void sortAllGroupsInOrderValueOrder(){
      for(int i=0; i<groupOrdered.size(); i++){
      IObjectGroup cog=groupOrdered.elementAt(i);
      cog.sortInOrderValueOrder();
      }
    */

    // Fljande kanske behver fixas senare
    /*
      objectOrderIndexMap = new HashMap<IObject, Integer>();
      for(int i=0; i<ordered.size(); i++){
      IObject co = ordered.elementAt(i);
      objectOrderIndexMap.put(co, i);
      }
    */
    //  }


    /**
     * Sorts all IObjectGroup:s in their internal order
     * and sorts the IObject:s in that order in the "ordered" vector.
     * 
     * @see IObjectGroup#sortInOrderValueOrder
     */
    public void sort(){
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup cog=groupOrdered.elementAt(i);
	    cog.sortInOrderValueOrder();
	}    	

	Vector<IObjectGroup> groups = getGroups();
	removeAll();
	for(int i=0; i<groups.size(); i++){
	    Vector<IObject> cobjects = groups.elementAt(i).getOrdered();
	    for(int j=0; j<cobjects.size(); j++){
		IObject co = cobjects.elementAt(j);
		add(co);
	    }
	}
	System.out.println("--------- objectOrderIndexMap: "+objectOrderIndexMap.size());

    }

    /**
     * Removes and returns the IObjects in the range given,
     * the range within the vector "ordered".
     */
    public Vector<IObject> remove(int from, int to){

	// Ta bort
	Vector<IObject> removed = new Vector<IObject>();
	for (int i=to;i>=from;i--){
	    IObject co = ordered.elementAt(i);
	    remove(co);
	    removed.add(co);
	}
	removeEmptyGroups();
		
	// Sortera om
	sort();
	/*
	  if(groupings!=null){
	  for(int i=0; i<groupings.size(); i++){
	  IObjectGrouping coging = groupings.elementAt(i);
	  coging.sortAllGroupsInOrderValueOrder();
	  }
	  }
	*/
	return removed;
    }


    /**
     * Removes IObject:s are not in the IObjectSet. 
     * Called by IObjectSet#prugeGroupingstoSet.
     *
     * @see IMatrixFilter, IObjectSet#prugeGroupingsToSet()
     */
    protected void purgeGroupingToSet(IObjectSet set){
	Vector<IObject> removed = new Vector<IObject>();
	for(int i=0; i<ordered.size(); i++){
	    IObject obj = ordered.elementAt(i);
	    if(!set.contains(obj))
		removed.add(obj);
	}
	for(int i=0; i<removed.size(); i++){
	    remove(removed.elementAt(i));
	}
	removeEmptyGroups();
	sort();
    }


    /**
     * Removes IObject:s that are not in the given IObjectGroup.
     * Does not check for the opposite: objects that are in this, but not
     * in the given.
     */
    public void removeObjectsNotIn(IObjectGroup group){
	// This has to be implemented here as well as in IObjectGroup,
	// since here they are also removed from the groups.
	Vector<IObject> removed = new Vector<IObject>();
	for(int i=0; i<ordered.size(); i++){
	    IObject obj = ordered.elementAt(i);
	    if(!group.contains(obj))
		removed.add(obj);
	}
	for(int i=0; i<removed.size(); i++){
	    remove(removed.elementAt(i));
	}
	removeEmptyGroups();
	sort();
    }



    /**
     * Returns a vector with integers that indicate where the boundaries
     * for the IObjectsGroups are in the entire ordered vector "groupOrdered".
     * The next integer is the first in the next group.
     * Used in SparsePictureConstructor.
     * @see infomat.view.picture.SparsePictureConstructor
     */
    public Vector<Integer> getGroupLimitsForIntervall(int from, int to){
	Vector<Integer> ret = new Vector<Integer>();
	Vector<IObject> all = getGroupingVector();

	// From
	IObject fromObject = all.elementAt(from);
	IObjectGroup fromGroup = inWhichIObjectGroup(fromObject);
	int fromGroupIndex = groupOrdered.indexOf(fromGroup);

	// To
	IObject toButOneObject = all.elementAt(to-1);
	IObjectGroup toGroup = inWhichIObjectGroup(toButOneObject);
	int toGroupIndex = groupOrdered.indexOf(toGroup);

	int now = from;
	ret.add(now);

	int f = fromGroup.getIObjectOrderIndex(fromObject);
	int s = fromGroup.numberOfIObjects();
	now += s - f;
	ret.add(now);

	if(toGroupIndex>fromGroupIndex){
	    for(int i=fromGroupIndex+1; i<toGroupIndex; i++){
		now += groupOrdered.elementAt(i).numberOfIObjects();
		ret.add(now);
	    }
	    ret.add(to);
	}

	return ret;
    }

    /**
     * Returns a vector of IObjectGroups that lie within the intervall given.
     * The intervall is within the ordered vector "groupOrdered".
     * Used in SparsePictureConstructor together with 
     * #getGroupLimitsForIntervall.
     * @see infomat.view.picture.SparsePictureConstructor
     */
    public Vector<IObjectGroup> getGroupsForIntervall(int from, int to){
	Vector<IObjectGroup> ret = new Vector<IObjectGroup>();
	Vector<IObject> all = getGroupingVector();

	// From
	IObject fromObject = all.elementAt(from);
	IObjectGroup fromGroup = inWhichIObjectGroup(fromObject);
	int fromGroupIndex = groupOrdered.indexOf(fromGroup);

	// To
	IObject toButOneObject = all.elementAt(to-1);
	IObjectGroup toGroup = inWhichIObjectGroup(toButOneObject);
	int toGroupIndex = groupOrdered.indexOf(toGroup);

	for(int i=fromGroupIndex; i<=toGroupIndex; i++){
	    ret.add(groupOrdered.elementAt(i));
	}
	return ret;
    }


    /**
     * For debugging purposes.
     */
    public void printGroupsObjects(int numberOfObjects){
	System.out.println("************** "+getName());
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup iog = groupOrdered.elementAt(i);
	    iog.sortInOrderValueOrder();
	    System.out.println("*** "+iog.getName());
	    System.out.print("Top objects: ");
	    Vector<IObject> ios = iog.getOrdered();
	    for(int j=0; j<numberOfObjects && j<ios.size(); j++){
		System.out.print(ios.elementAt(j).getString());
		if(j<ios.size()-1 && j+1<numberOfObjects)
		    System.out.print(", ");
	    }
	    System.out.println();
	}
    }



    /**
     * For debugging purposes.
     */
    /*
    public void printGroupsRepresentationObjects(int numberOfObjects, IMatrix matrix){
	System.out.println("************** "+getName());
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup iog = groupOrdered.elementAt(i);
	    
	    Vector<Map.Entry<IObject, Float>> reprObjects = 
		iog.getSortedCentroid(matrix);
	    System.out.println("*** "+iog.getName());
	    System.out.print("Top representation objects: ");
	    for(int j=0; j<numberOfObjects && j<reprObjects.size(); j++){
		Map.Entry<IObject, Float> entry = reprObjects.elementAt(j);
		System.out.print(entry.getKey().getString());
		if(j<reprObjects.size()-1 && j+1<numberOfObjects)
		    System.out.print(", ");
	    }
	    System.out.println();
	}
    }
    */

    /**
     * For debugging purposes.
     */
    public void printGroupNamesAndSizes(){
	for(int i=0; i<groupOrdered.size(); i++){
	    IObjectGroup group = groupOrdered.elementAt(i);
	    
	    System.out.print(group.getName()+": ");
	    System.out.println(group.numberOfIObjects());
	}
    }

    /**
     * For debugging purposes.
     */
    public void printGroupSizes(){
	for(int i=0; i<groupOrdered.size(); i++){
	    System.out.print(groupOrdered.elementAt(i).numberOfIObjects()+" ");
	}
	System.out.println();
    }


}

